#include "SimpleTransceiver.h"
//-------------------------------------------------------------------------------------------------
CSimpleTransceiver::CSimpleTransceiver()
{
	m_shutdown = false;

	cout << "CSimpleTransceiver constructor called." << endl;

	m_pMutexRxQueue = new boost::mutex;
	m_pMutexTxQueue = new boost::mutex;

	m_pTxQueue = new list<vector<complex<Float32> >*>;
	m_pRxQueue = new list<vector<complex<Float32> >*>;

	setupUsrp();

	// Start up the threads
	m_pRxThread = new boost::thread(&CSimpleTransceiver::rxThread,this);
	m_pTxThread = new boost::thread(&CSimpleTransceiver::txThread,this);
	m_pProcessingThread = new boost::thread(&CSimpleTransceiver::processingThread,this);
}
//-------------------------------------------------------------------------------------------------
CSimpleTransceiver::~CSimpleTransceiver()
{
	cout << "CTransceiver destructor called." << endl;
	m_shutdown = true;
	m_pTxThread->join();
	m_pRxThread->join();
	m_pProcessingThread->join();

	// cleanup
	delete m_pTxThread;
	delete m_pRxThread;
	delete m_pProcessingThread;
	delete m_pMutexTxQueue;
	delete m_pTxQueue;
	delete m_pMutexRxQueue;
	delete m_pRxQueue;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::txThread()
{
	// TODO make configurable
	const Int32 sleepTimeMs = 1000;

	cout << "CSimpleTransceiver::txThread() started." << endl;

	list<vector<complex<Float32> >* > tmpList;
	list<vector<complex<Float32> >* >::iterator it;

	while (!m_shutdown)
	{
		// Lock queue
		{
			 boost::mutex::scoped_lock l(*m_pMutexTxQueue);
			 tmpList.clear();
			// See if there are messages on queue
			 Int32 qSize = m_pTxQueue->size();
			 if (qSize>0)
			 {
				 // Remove messages from the queue so that they can go to processing.
				 tmpList.assign(m_pTxQueue->begin(), m_pTxQueue->end());
			 }
			 m_pTxQueue->clear();
		}

		// Process the messages
		Int32 s = tmpList.size();
		if (s>0)
		{
			it = tmpList.begin();
			for (Int32 k=0;k<s;++k)
			{
				// Here is where modulation could be done.

				// Send the samps over the air.
				sendSampsOverRf(*it);

				// Cleanup
				delete (*it);
				++it;
			}

			stringstream msg;
				msg << "CSimpleTransceiver::txThread() sent " << s << " messages.";
				cout << msg.str() << endl;
		}
		// Sleep
		boost::this_thread::sleep(boost::posix_time::milliseconds(sleepTimeMs));
	}

	cout << "CSimpleTransceiver::txThread() exited." << endl;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::sendSampsOverRf(vector<complex<Float32> >* pVec)
{
	// Set up tx params.  We wish to send a burst.
	uhd::tx_metadata_t md;
    md.start_of_burst = true;
    md.end_of_burst   = true;
    md.has_time_spec  = false;

	//send the entire packet (driver fragments internally)
	size_t num_tx_samps = m_pDev->send(&pVec->front(), pVec->size(), md,uhd::io_type_t::COMPLEX_FLOAT32,uhd::device::SEND_MODE_FULL_BUFF);

    std::cout << std::endl << boost::format("Sent %d samples") % num_tx_samps << std::endl;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::rxThread()
{
	cout << "CTransceiver::rxThread() started." << endl;

	// Allocation and init
	uhd::rx_metadata_t md;
	vector<complex<Float32> >* pBuff = NULL;
	stringstream msg;

	while (!(m_shutdown))
	{
		msg.str("");

		// Rx data until the buffer (data block) is full.  Buffer len is m_usrpSampsPerBlock.
		pBuff = new vector<complex<Float32> >(m_usrpSampsPerBlockRx);

		// Rx the data
		size_t num_rx_samps = m_pDev->recv(&(pBuff->front()),pBuff->size(),md,uhd::io_type_t::COMPLEX_FLOAT32,uhd::device::RECV_MODE_FULL_BUFF);

		// Some checks.
		if (m_usrpSampsPerBlockRx != num_rx_samps)
		{
			cout << "m_usrpSampsPerBlock != num_rx_samps!!!" << endl;
		}
		if (0 == num_rx_samps)
		{
			cout << "No samps rx." << endl;
		}

		//handle the error codes
		// idg - I believe the timeout and overflow codes dont work for USRP1.
		switch(md.error_code)
		{
			case uhd::rx_metadata_t::ERROR_CODE_NONE:
				goto done_loop;
			case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
				msg << "Got timeout before all samples received, possible packet loss, exiting loop...";
				cout << msg.str() << endl;
				goto done_loop;
			case uhd::rx_metadata_t::ERROR_CODE_OVERFLOW:
				msg << "Overflow!";
				cout << msg.str() << endl;
				goto done_loop;
			default:
				cout << "Got error code 0x"<< md.error_code << ", exiting loop..." << endl;
				goto done_loop;
		}

	    done_loop:

	    // Add the datablock to the queue.
		{
			 // Lock queue
	    	 boost::mutex::scoped_lock l(*m_pMutexRxQueue);
			 // Put data block on queue
			 m_pRxQueue->push_back(pBuff);
		}	// Unlock queue
	}

	cout << "Shutting down the USRP." << endl;
	// Gracefully shutdown the radio
	// TODO maybe put this somewehere else?  have an elegant shutdown function.
	(m_pSdev)->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS);

	cout << "CTransceiver::rxThread() exited." << endl;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::processBlock(vector<complex<Float32> >* pBlock)
{
	// Simple energy detector
	// Prints out a 1 if the energy present at the sample, 0 otherwise.

	const Float32 thresh = 0.0001;

	Bool flag = false;
	Float64 v = 0.0;
	const Int32 printBlockWidth = 50;

	for (Int32 k=0;k<pBlock->size();++k)
	{
		v = ((*pBlock)[k]).real()*((*pBlock)[k]).real() + ((*pBlock)[k]).imag()*((*pBlock)[k]).imag();
		if (v > thresh)
		{
			flag = true;
			break;
		}
	}

	// Print block out if energy found.
	if (flag)
	{
		for (Int32 k=0;k<pBlock->size();++k)
		{
			v = ((*pBlock)[k]).real()*((*pBlock)[k]).real() + ((*pBlock)[k]).imag()*((*pBlock)[k]).imag();
			cout << (v > thresh) << " ";
			if (!((k+1) % printBlockWidth)) cout << endl;
		}
		cout << endl;
	}

	// Clean up after ourselves.
	delete pBlock;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::setupUsrp()
{
	cout << "CSimpleTransceiver::setupUsrp() called." << endl;

	// TODO This fails, maybe has to run as root?
	uhd::set_thread_priority_safe();

	// Adapted from:
	// http://ettus-apps.sourcerepo.com/redmine/ettus/projects/uhd/repository/revisions/master/entry/host/examples/rx_samples_to_file.cpp
	// http://ettus-apps.sourcerepo.com/redmine/ettus/projects/uhd/repository/revisions/master/entry/host/examples/rx_ascii_art_dft.cpp

    String args = String("");
    m_pSdev = uhd::usrp::single_usrp::make(args);

    // Set exactly what you want to use
    //m_pSdev->set_rx_subdev_spec(uhd::usrp::subdev_spec_t("A:AB"));
    //m_pSdev->set_tx_subdev_spec(uhd::usrp::subdev_spec_t("A:AB"));

    m_pDev = m_pSdev->get_device();
    std::cout << boost::format("Using Device: %s") % m_pSdev->get_pp_string() << std::endl;

    //set the rx sample rate
    Float64 rate = 64e6/(256);
    std::cout << boost::format("Setting RX Rate: %f sps...") % (rate) << std::endl;
    m_pSdev->set_rx_rate(rate);
    std::cout << boost::format("Actual RX Rate: %f sps...") % (m_pSdev->get_rx_rate()) << std::endl << std::endl;

    //set the rx center frequency
    Float64 freq = 1e6;
    std::cout << boost::format("Setting RX Freq: %f Mhz...") % (freq/1e6) << std::endl;
    m_pSdev->set_rx_freq(freq);
    std::cout << boost::format("Actual RX Freq: %f Mhz...") % (m_pSdev->get_rx_freq()/1e6) << std::endl << std::endl;

    //set the rx rf gain
    Float64 gain = 0;
    std::cout << boost::format("Setting RX Gain: %f dB...") % gain << std::endl;
    m_pSdev->set_rx_gain(gain);
    std::cout << boost::format("Actual RX Gain: %f dB...") % m_pSdev->get_rx_gain() << std::endl << std::endl;

    // Setup the Tx path
    //set the tx sample rate
    std::cout << boost::format("Setting TX Rate: %f sps...") % (rate) << std::endl;
    m_pSdev->set_tx_rate(rate);
    std::cout << boost::format("Actual TX Rate: %f sps...") % (m_pSdev->get_tx_rate()) << std::endl << std::endl;

    //set the tx center frequency
    std::cout << boost::format("Setting TX Freq: %f Mhz...") % (freq/1e6) << std::endl;
    m_pSdev->set_tx_freq(freq);
    std::cout << boost::format("Actual TX Freq: %f Mhz...") % (m_pSdev->get_tx_freq()/1e6) << std::endl << std::endl;

    //set the tx rf gain
    std::cout << boost::format("Setting TX Gain: %f dB...") % gain << std::endl;
    m_pSdev->set_tx_gain(gain);
    std::cout << boost::format("Actual TX Gain: %f dB...") % m_pSdev->get_tx_gain() << std::endl << std::endl;

    // Next 2 lines not present with the ascii art example.
    boost::this_thread::sleep(boost::posix_time::seconds(1)); //allow for some setup time
    std::cout << "LO Locked = " << m_pSdev->get_rx_lo_locked() << std::endl;

    m_usrpSampsPerBlockRx = m_pDev->get_max_recv_samps_per_packet();

    // http://ettus-apps.sourcerepo.com/redmine/ettus/projects/uhd/repository/revisions/master/entry/host/examples/benchmark_rx_rate.cpp
    // Flush rx buffer
    // idg - This seems to do nothing.
    //
    uhd::rx_metadata_t md;
    std::vector<std::complex<Float32> > buff(m_usrpSampsPerBlockRx);
    while(m_pDev->recv(&buff.front(), buff.size(), md,uhd::io_type_t::COMPLEX_FLOAT32,uhd::device::RECV_MODE_ONE_PACKET))
    {
    	// NOP
    }
	//handle the error codes
    stringstream msg;
	switch(md.error_code)
	{
		case uhd::rx_metadata_t::ERROR_CODE_NONE:
			cout << "USRP flush OK" << endl;
			break;
		case uhd::rx_metadata_t::ERROR_CODE_TIMEOUT:
			//if (num_acc_samps == 0) continue;
			cout << "Got timeout before all samples received, possible packet loss, exiting loop..." << endl;
			break;
		case uhd::rx_metadata_t::ERROR_CODE_OVERFLOW:
			//if (num_acc_samps == 0) continue;
			cout << "Overflow!" << endl;
			break;
		default:
			cout << "Got error code 0x"<< md.error_code << ", exiting loop..." << endl;
			break;
	}

    //setup rx streaming
    m_pSdev->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS);

    cout << "CTransceiver::setupUsrp() setup complete." << endl;
}
//-------------------------------------------------------------------------------------------------

void CSimpleTransceiver::processingThread()
{
	cout << "CSimpleTransceiver::rxQueueProcessingThread started." << endl;

	list<vector<complex<Float32> >*> tmpBlockList;
	list<vector<complex<Float32> >*>::iterator it;

	stringstream msg;
	Int32 qSize = 0;

	const Int32 sleepTimeMs = 100;

	while (!(m_shutdown && (0==m_pRxQueue->size())))
	{
		if (m_shutdown)
		{
			cout << "Shutdown observed.  Working off remaining blocks..." << endl;
		}

		// Lock queue
		{
			 boost::mutex::scoped_lock l(*m_pMutexRxQueue);

			 // Pull all the data from the queue
			 tmpBlockList.clear();
			 qSize = m_pRxQueue->size();
			 for (Int32 k=0;k<qSize;++k)
			 {
				 tmpBlockList.push_back(m_pRxQueue->front());
				 m_pRxQueue->pop_front();
			 }
		} // Unlock queue

		Int32 numItems = tmpBlockList.size();
		// Process each block of data
		it = tmpBlockList.begin();
		for (Int32 k=0;k<numItems;++k)
		{
			processBlock(*it);
			++it;
		}
		boost::this_thread::sleep(boost::posix_time::milliseconds(sleepTimeMs));
	}

	cout << "CSimpleTransceiver::processingThread exiting." << endl;
}

//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::postWaveformToTxQueue(vector<complex<Float32> >* pVec)
{
	{
		// Lock queue
		boost::mutex::scoped_lock l(*m_pMutexTxQueue);
		// Add the data block
		m_pTxQueue->push_back(pVec);
	}

	cout << "Posted waveform to Tx queue." << endl;
}
//-------------------------------------------------------------------------------------------------
void CSimpleTransceiver::postBurst(Int32 burstLenSamps)
{
	const Float32 gain = 1.0;
	vector<complex<Float32> >* pVec = new vector<complex<Float32> >(burstLenSamps, complex<Float32>(gain, gain));

	postWaveformToTxQueue(pVec);
}
//-------------------------------------------------------------------------------------------------

