///
/// Simple Transceiver class
///
/// There are 2 queues in this class. One for tx messages and one for the
/// the rx'ed data blocks.
///
/// There are three threads in this class:
/// 1. rxThread - rx's blocks of data and puts them in a queue for processing.
/// 2. txThread - checks the tx queue peridocially for any messages and sends them.
/// 3. processingThread - wakes up periodically and processes the rx'ed data blocks.
///
/// Author: Isaac Gerg
///

#ifndef _SIMPLETRANSCEIVER_H_
#define _SIMPLETRANSCEIVER_H_

// System includes
#include <list>
#include <string>
#include <complex>
#include <iostream>
#include <sstream>
#include <math.h>

// Boost includes
#include <boost/thread.hpp>
#include <boost/date_time.hpp>

// USRP includes
#include <uhd/utils/thread_priority.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/usrp/single_usrp.hpp>

using namespace std;

// Project includes
#include "typedefs.h"

class CSimpleTransceiver
{
public:
	/// Default constructor
	CSimpleTransceiver();
	virtual ~CSimpleTransceiver();
	
	/// Posts a burst waveform to the tx queue.
	/// @param burstLenSamps Length of theh burst to send in samples.
	void postBurst(Int32 burstLenSamps);

private:
	// Tx's data that is on the queue
	void txThread();
	// Rx's data and put it on the queue
	void rxThread();
	// Checks to see if there are data blocks on the queue and then processes them.
	void processingThread();

	// Posts the waveform to the Tx queue
	void postWaveformToTxQueue(vector<complex<Float32> >* pVec);
	
	// Actually sends the samples over the air.
	void sendSampsOverRf(vector<complex<Float32> >* pVec);

	// Sets up the USRP
	void setupUsrp();

	// Processes a rx block of data
	void processBlock(vector<complex<Float32> >* pBlock);

	Bool m_shutdown;

	// Threads
	boost::thread* m_pTxThread;
	boost::thread* m_pRxThread;
	boost::thread* m_pProcessingThread;

	// Queues
	boost::mutex* m_pMutexTxQueue;
	list<vector<complex<Float32> >*>* m_pTxQueue;
	boost::mutex* m_pMutexRxQueue;
	list<vector<complex<Float32> >* >* m_pRxQueue;

	// USRP
	uhd::device::sptr m_pDev;
	uhd::usrp::single_usrp::sptr m_pSdev;
	Int32 m_usrpSampsPerBlockRx;
};

#endif /* _SIMPLETRANSCEIVER_H_ */
